//==========================================================================
//
//      netarm_eth_drv.c
//
//      NetSilion NET+ARM Ethernet Driver (DMA driven)
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System
// Copyright (C) 2005 eCosCentric Ltd.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):           Harald Brandl (harald.brandl@fh-joanneum.at)
// Contributors:        Harald Brandl
// Date:                01.08.2004
// Purpose:             NET+ARM Ethernet Driver (DMA driven)
// Description:
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/devs_eth_arm_netarm.h>

#include <cyg/hal/hal_platform_ints.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/drv_api.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/infra/diag.h>
#include <cyg/io/eth/netdev.h>
#include <cyg/io/eth/eth_drv.h>

#include "eth_regs.h"
#include "MII.h"
#include "netarm_eth_drv.h"

#define CYGNUM_HAL_INTERRUPT_DMA2               31
#define CYGNUM_HAL_INTERRUPT_DMA1               32

#define BufferSizeA 128
#define BufferSizeB 512
#define BufferSizeC 1500

#define NumA 10
#define NumB 5
#define NumC 10

static cyg_mutex_t Key_Mutex;

static private_data_t driver_private = {NULL, 0, 0};

static unsigned char esa[6] = CYGDAT_DEVS_ETH_ARM_NETARM_ETH0_ESA;

static unsigned char RxBufferA[NumA][BufferSizeA];
static unsigned char RxBufferB[NumB][BufferSizeB];
static unsigned char RxBufferC[NumC][BufferSizeC];

static unsigned char TxBuffer[1500];
static unsigned char TxHeader[14];

static BDP_t RxBDP_A[NumA];
static BDP_t RxBDP_B[NumB];
static BDP_t RxBDP_C[NumC];

static BDP_t TxBDP[2];

ETH_DRV_SC(netarm_sc,
           (void *)&driver_private,     // driver specific data
           "eth0",                      // Name for this interface
           netarm_start,
           netarm_stop,
           netarm_control,
           netarm_can_send,
           netarm_send,
           netarm_recv,
           netarm_deliver,
           netarm_poll,
           netarm_int_vector);

NETDEVTAB_ENTRY(netarm_netdev,
                "ETH_DRV",
                netarm_init,
                &netarm_sc);

static void 
fastcopy(void *buf, void *data, unsigned long len)
{
  asm volatile(

               "STMDB SP!, {R11};"

               "TST R1, #2;"                   // test if aligned
               "LDRNEH R3, [R1], #2;"
               "STRNEH R3, [R0], #2;"
               "SUBNE R2, R2, #2;"
               "TST R1, #1;"
               "LDRNEB R3, [R1], #1;"
               "STRNEB R3, [R0], #1;"
               "SUBNE R2, R2, #1;"

               ".START:"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "CMP R2, #44;"
               "BLT .LASTBYTES;"
               "LDMIA R1!, {R3 - R12, R14};"
               "STMIA R0!, {R3 - R12, R14};"
               "SUB R2, R2, #44;"

               "BGE .START;"


               ".LASTBYTES:"

               "AND R14, R2, #0xfffffffc;"
               "LDR PC, [PC, R14];"
               "NOP;"

               ".SWITCH:"
               ".word  .CASE0;"
               ".word  .CASE1;"
               ".word  .CASE2;"
               ".word  .CASE3;"
               ".word  .CASE4;"
               ".word  .CASE5;"
               ".word  .CASE6;"
               ".word  .CASE7;"
               ".word  .CASE8;"
               ".word  .CASE9;"
               ".word  .CASE10;"

               ".CASE0:"
               "B .END;"

               ".CASE1:"
               "LDR R3, [R1]!;"
               "STR R3, [R0]!;"
               "B .END;"

               ".CASE2:"
               "LDMIA R1!, {R3, R4};"
               "STMIA R0!, {R3, R4};"
               "B .END;"

               ".CASE3:"
               "LDMIA R1!, {R3 - R5};"
               "STMIA R0!, {R3 - R5};"
               "B .END;"

               ".CASE4:"
               "LDMIA R1!, {R3 - R6};"
               "STMIA R0!, {R3 - R6};"
               "B .END;"

               ".CASE5:"
               "LDMIA R1!, {R3 - R7};"
               "STMIA R0!, {R3 - R7};"
               "B .END;"

               ".CASE6:"
               "LDMIA R1!, {R3 - R8};"
               "STMIA R0!, {R3 - R8};"
               "B .END;"

               ".CASE7:"
               "LDMIA R1!, {R3 - R9};"
               "STMIA R0!, {R3 - R9};"
               "B .END;"

               ".CASE8:"
               "LDMIA R1!, {R3 - R10};"
               "STMIA R0!, {R3 - R10};"
               "B .END;"

               ".CASE9:"
               "LDMIA R1!, {R3 - R11};"
               "STMIA R0!, {R3 - R11};"
               "B .END;"

               ".CASE10:"
               "LDMIA R1!, {R3 - R12};"
               "STMIA R0!, {R3 - R12};"

               ".END:"
               "TST R2, #2;"
               "LDRNEH R3, [R1], #2;"
               "STRNEH R3, [R0], #2;"
               "TST R2, #1;"
               "LDRNEB R3, [R1], #1;"
               "STRNEB R3, [R0], #1;"

               "LDMIA SP!, {R11};"

               :
               : "r" (buf), "r" (data), "r" (len)
               : "r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11", "r12", "r14"
               );
}

static bool 
KeyBufferFull(private_data_t *pd)
{
     int tmp = 0;
     
     cyg_drv_mutex_lock(&Key_Mutex);
     
     if((pd->key_head + 1) % MaxKeys == pd->key_tail)
     {
          tmp = 1;
     }
     
     cyg_drv_mutex_unlock(&Key_Mutex);
     
     return tmp;
}


static void 
AddKey(unsigned long key, private_data_t *pd)
{

     cyg_drv_mutex_lock(&Key_Mutex);

     pd->KeyBuffer[pd->key_head] = key;
     pd->key_head = (pd->key_head + 1) % MaxKeys;

     cyg_drv_mutex_unlock(&Key_Mutex);
}

static unsigned long 
GetKey(private_data_t *pd)
{
     unsigned long key = 0;

     cyg_drv_mutex_lock(&Key_Mutex);

     if(pd->key_tail != pd->key_head)
     {
          key = pd->KeyBuffer[pd->key_tail];
          pd->key_tail = (pd->key_tail + 1) % MaxKeys;
     }

     cyg_drv_mutex_unlock(&Key_Mutex);

     return key;
}


static cyg_uint32 
dma_rx_isr(cyg_vector_t vector, cyg_addrword_t data)
{

     // block this interrupt until the dsr completes
     cyg_drv_interrupt_mask(vector);

     // tell ecos to allow further interrupt processing
     cyg_drv_interrupt_acknowledge(vector);

     return CYG_ISR_CALL_DSR;        // call the dsr
}

static void 
dma_rx_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{

     eth_drv_dsr( vector, count, data );
     
     DMA1A_SR |= 0x80000000;     // acknowledge normal completion interrupt
     DMA1B_SR |= 0x80000000;
     DMA1C_SR |= 0x80000000;
     
     cyg_drv_interrupt_unmask(vector);
}

static cyg_uint32 
dma_tx_isr(cyg_vector_t vector, cyg_addrword_t data)
{

     // block this interrupt until the dsr completes
     cyg_drv_interrupt_mask(vector);

     // tell ecos to allow further interrupt processing
     cyg_drv_interrupt_acknowledge(vector);

     return CYG_ISR_CALL_DSR;        // invoke the dsr
}

static void 
dma_tx_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{

     eth_drv_dsr( vector, count, data );

     DMA2_SR |= 0x80000000; // acknowledge interrupt

     cyg_drv_interrupt_unmask(vector);
}

static void 
SetupDMA(void)
{
     int i;

     for(i = 0; i < NumA; i++)
     {
          RxBDP_A[i].BufferPointer = ((unsigned long)RxBufferA[i]) & 0x3fffffff;
          RxBDP_A[i].BufferLen = BufferSizeA;
     }

     RxBDP_A[i - 1].W = 1;

     for(i = 0; i < NumB; i++)
     {
          RxBDP_B[i].BufferPointer = ((unsigned long)RxBufferB[i]) & 0x3fffffff;
          RxBDP_B[i].BufferLen = BufferSizeB;
     }

     RxBDP_B[i - 1].W = 1;

     for(i = 0; i < NumC; i++)
     {
          RxBDP_C[i].BufferPointer = ((unsigned long)RxBufferC[i]) & 0x3fffffff;
          RxBDP_C[i].BufferLen = BufferSizeC;
     }

     RxBDP_C[i - 1].W = 1;

     DMA1A_BDP = (unsigned long)RxBDP_A;
     DMA1A_CR = 0x82000000;  //burst transfer
     DMA1A_SR |= 0xa00000;

     DMA1B_BDP = (unsigned long)RxBDP_B;
     DMA1B_CR = 0x82000000;  //burst transfer
     DMA1B_SR |= 0xa00000;

     DMA1C_BDP = (unsigned long)RxBDP_C;
     DMA1C_CR = 0x82000000;  //burst transfer
     DMA1C_SR |= 0xa00000;


     // setup dma transmitter

     TxBDP[0].BufferPointer = ((unsigned long)TxHeader) & 0x3fffffff;
     TxBDP[1].BufferPointer = ((unsigned long)TxBuffer) & 0x3fffffff;
     TxBDP[1].L = 1;
     TxBDP[1].W = 1;

     DMA2_BDP = (unsigned long)TxBDP;
     DMA2_CR = 0x86000000;   //burst transfer
     DMA2_SR |= 0x800000;

}


static bool 
netarm_init(struct cyg_netdevtab_entry *tab)
{
     struct eth_drv_sc *sc = (struct eth_drv_sc *)tab->device_instance;
     cyg_bool duplex;
     static cyg_interrupt dma_rx_int_object, dma_tx_int_object;
     static cyg_handle_t dma_rx_int_handle, dma_tx_int_handle;

     // setup dma receiver
     SetupDMA();

     cyg_netarm_mii_reset();
     cyg_netarm_mii_negotiate();    // initialize PHY
     duplex = cyg_netarm_mii_check_duplex();

     // Ethernet Controller Initialisation

     // auto CRC, late collision retry
     MACCR  = (0x1c | (duplex << 1));            
     // insert MAC source address into ethernet frame
     STLCR = 0x3; 
     BtBIPGGapTimerR = 0x14;                     // standard values
     NonBtBIPGGapTimerR = ((0x9 << 7) | 0x11);   // standard values
     CollWinR = ((0x37 << 8) | 0xf);             // standard values
     SAFR = 0x1;                                 // broadcast mode
     // dma mode, full duplex, enable pNA mode(needed for alignment)
     EthGenCR = (0x40400400 | (duplex << 16));   

     cyg_drv_interrupt_create(
          CYGNUM_HAL_INTERRUPT_DMA1,  // Interrupt Vector
          0,                          // Interrupt Priority
          (cyg_addrword_t)&netarm_sc, // Reference to Driver Instance
          dma_rx_isr,
          dma_rx_dsr,
          &dma_rx_int_handle,
          &dma_rx_int_object);

     cyg_drv_interrupt_create(
          CYGNUM_HAL_INTERRUPT_DMA2,  // Interrupt Vector
          0,                          // Interrupt Priority
          (cyg_addrword_t)&netarm_sc, // Reference to Driver Instance
          dma_tx_isr,
          dma_tx_dsr,
          &dma_tx_int_handle,
          &dma_tx_int_object);

     cyg_drv_interrupt_attach(dma_rx_int_handle);
     cyg_drv_interrupt_attach(dma_tx_int_handle);

     cyg_mutex_init(&Key_Mutex);

     sc->funs->eth_drv->init(sc, esa);

     return true;
}


static void 
netarm_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len)
{
     int i, len;
     private_data_t *pd = (private_data_t *)(sc->driver_private);
     unsigned char *data, *buf = pd->RxBuffer;

     for (i = 0;  i < sg_len;  i++) {
          data = (unsigned char *)(sg_list[i].buf);
          len = sg_list[i].len;
          if (len)
          {
               if(i == 1)
               {
                    buf += 2;
               }

               fastcopy(data, buf, len);
               buf += len;
          }
     }
}


static void 
netarm_deliver(struct eth_drv_sc *sc)
{
     static int a = 0, b = 0, c = 0;
     unsigned long key;
     private_data_t *pd = (private_data_t *)(sc->driver_private);

     while((key = GetKey(pd)))
     {
          sc->funs->eth_drv->tx_done(sc, key, 0);
     }

     while(RxBDP_A[a].F)
     {
          pd->RxBuffer = (unsigned char *)RxBDP_A[a].BufferPointer;
          sc->funs->eth_drv->recv(sc, RxBDP_A[a].BufferLen);
          RxBDP_A[a].BufferLen = BufferSizeA;
          RxBDP_A[a].F = 0;
          a = (a + 1) % NumA;
     }

     while(RxBDP_B[b].F)
     {
          pd->RxBuffer = (unsigned char *)RxBDP_B[b].BufferPointer;
          sc->funs->eth_drv->recv(sc, RxBDP_B[b].BufferLen);
          RxBDP_B[b].BufferLen = BufferSizeB;
          RxBDP_B[b].F = 0;
          b = (b + 1) % NumB;
     }

     while(RxBDP_C[c].F)
     {
          pd->RxBuffer = (unsigned char *)RxBDP_C[c].BufferPointer;
          sc->funs->eth_drv->recv(sc, RxBDP_C[c].BufferLen);
          RxBDP_C[c].BufferLen = BufferSizeC;
          RxBDP_C[c].F = 0;
          c = (c + 1) % NumC;
     }

     cyg_drv_dsr_lock();
     DMA1A_SR |= 0x20000000;
     DMA1B_SR |= 0x20000000;
     DMA1C_SR |= 0x20000000;
     cyg_drv_dsr_unlock();
}


static int 
netarm_can_send(struct eth_drv_sc *sc)
{
     private_data_t *pd = (private_data_t *)(sc->driver_private);

     if(TxBDP[0].F || TxBDP[1].F || KeyBufferFull(pd))
     {
          return 0;
     }

     return 1;
}

static void 
netarm_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, 
            int sg_len, int total_len, unsigned long key)
{
     private_data_t *pd = (private_data_t *)(sc->driver_private);
     register int i, len;
     register unsigned char *data, *buf = TxBuffer;
     
     AddKey(key, pd);
     
     // Put data into buffer
     for (i = 0;  i < sg_len;  i++) {
          data = (unsigned char *)sg_list[i].buf;
          len = sg_list[i].len;
          
          if(i == 0)
          {
               if((unsigned long)data & 0x3)
               {
                    memcpy(TxHeader, data, 0xe);
               }
               else
               {
                    fastcopy(TxHeader, data, 0xe);
               }
               
               len -= 0xe;
               data += 0xe;
          }
          
          if (len)
          {
               fastcopy(buf, data, len);
               buf += len;
          }
     }
     
     TxBDP[0].BufferLen = 0xe;
     TxBDP[0].F = 1;
     TxBDP[1].BufferLen = total_len - 0xe;
     TxBDP[1].F = 1;
     
     cyg_drv_dsr_lock();
     DMA2_SR |= 0xf0000000;
     cyg_drv_dsr_unlock();
}


static void 
netarm_start(struct eth_drv_sc *sc, unsigned char *enaddr, int flags)
{
     SetMAC(enaddr);                 // set MAC address
     EthGenCR |= 0x80800000; // enable Rx und Tx FIFO
     cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_DMA1);
     cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_DMA2);
}

static void 
netarm_stop(struct eth_drv_sc *sc)
{
     EthGenCR &= 0x7f7fffff;
     cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_DMA1);
     cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_DMA2);
}

static int 
netarm_control(struct eth_drv_sc *sc, unsigned long key, void *data, int len)
{
     switch (key)
     {
     case ETH_DRV_SET_MAC_ADDRESS:
          return 0;
          break;
     case ETH_DRV_SET_MC_LIST:
     case ETH_DRV_SET_MC_ALL:
          return 0;
     default:
          return 1;
          break;
     }
}

static void 
netarm_poll(struct eth_drv_sc *sc)
{
}

static int 
netarm_int_vector(struct eth_drv_sc *sc)
{
        return CYGNUM_HAL_INTERRUPT_DMA1;
}

static void 
SetMAC(unsigned char *MAC)
{
     strcpy(esa, MAC);
     SAR1 = (esa[1] << 8) | esa[0];  // set MAC address
     SAR2 = (esa[3] << 8) | esa[2];
     SAR3 = (esa[5] << 8) | esa[4];
}
